#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

/* ビットマップファイル形式用ヘッダ */
#include "bitmap.h"

/* 読み込み可能なビットマップファイルの条件(固定値) */
#define BMP_PLANES      1  /* プレーン数 */
#define BMP_BITCOUNT    24 /* 色深度 */
#define BMP_COMPRESSION 0  /* 圧縮方式 */

/* 読み込む最大画素サイズ */
#define MAX_WIDTH    480 /* 幅 */
#define MAX_HEIGHT   272 /* 高さ */
/* キャラクタ化する画素単位 */
#define PIXEL_WIDTH  8  /* 幅 */
#define PIXEL_HEIGHT 16 /* 高さ */
/* 表示するキャラクタ個数 */
#define DISP_WIDTH   (MAX_WIDTH / PIXEL_WIDTH)   /* 幅 */
#define DISP_HEIGHT  (MAX_HEIGHT / PIXEL_HEIGHT) /* 高さ */

/* 色変換用境界値 */
#define BOUNDARY_FG (0x55 * PIXEL_WIDTH * PIXEL_HEIGHT) /* 前景 */
#define BOUNDARY_BG (0xaa * PIXEL_WIDTH * PIXEL_HEIGHT) /* 背景 */

/* 色変換用構造体 */
/* 同一キャラクタに当たる複数の画素について各色を加算していき、
   境界値と比較して色変換を行う */
typedef struct {
	uint16_t r; /* 赤 */
	uint16_t g; /* 緑 */
	uint16_t b; /* 青 */
} color_t;

#define BASENAME(p)   ((strrchr((p), '/') ? : ((p) - 1)) + 1)

/**
 * main関数
 * @param argc 引数なしの場合はusage表示のみ
 * @param argv 第1引数として読み込みビットマップファイル名を指定
 * @return exit値
 */
int main(int argc, char *argv[])
{
	FILE *pbmpfile; /* ビットマップファイルポインタ */
	BITMAPFILEHEADER bf; /* ビットマップファイルヘッダ */
	uint32_t offbits; /* ファイル先頭から画素データへのオフセットサイズ */
	BITMAPINFOHEADER bi; /* ビットマップ情報ヘッダ */
	int32_t bmp_width, bmp_height; /* ビットマップファイル画素サイズ */
	color_t pixels[DISP_HEIGHT][DISP_WIDTH]; /* 色変換用画素データ蓄積 */
	uint8_t buf[MAX_WIDTH][3], *ptmp; /* 画素データ読み込み用バッファ */
	int dx, dy, px, py; /* 画素シーク位置 */

	exitfail_init();

	/* 引数が指定されなかった場合、usage表示して終了 */
	if (argc < 2) {
		printf("Usage: %s <bmpfile>\n", BASENAME(argv[0]));
		return 0;
	}

	/* ビットマップファイルオープン */
	pbmpfile = fopen(argv[1], "r");
	if (!pbmpfile)
		exitfail_errno("fopen");

	/* ビットマップファイルヘッダ読み込み */
	if (fread(&bf, sizeof(bf), 1, pbmpfile) < 1)
		exitfail_errno("fread");
	/* ビットマップファイルタイプが正しい識別子か */
	if (BITMAPFILEHEADER_TYPE(&bf) != BF_TYPE)
		exitfail("Bad bitmap file header type\n");
	/* ビットマップヘッダサイズが十分か */
	offbits = BITMAPFILEHEADER_OFFBITS(&bf);
	if (offbits < sizeof(bf) + sizeof(bi))
		exitfail("Bad bitmap header size\n");

	/* ビットマップ情報ヘッダ読み込み */
	if (fread(&bi, sizeof(bi), 1, pbmpfile) < 1)
		exitfail_errno("fread");
	/* ビットマップ情報ヘッダサイズが十分か */
	if (BITMAPINFOHEADER_SIZE(&bi) < sizeof(bi))
		exitfail("Bad bitmap info header size\n");
	/* 読み込み可能なビットマップファイルか */
	if (BITMAPINFOHEADER_PLANES(&bi) != BMP_PLANES ||
	    BITMAPINFOHEADER_BITCOUNT(&bi) != BMP_BITCOUNT ||
	    BITMAPINFOHEADER_COMPRESSION(&bi) != BMP_COMPRESSION)
		exitfail("Bad bitmap type\n");
	/* ビットマップサイズが十分か */
	bmp_width = BITMAPINFOHEADER_WIDTH(&bi);
	bmp_height = BITMAPINFOHEADER_HEIGHT(&bi);
	if (bmp_width < MAX_WIDTH || bmp_height < MAX_HEIGHT)
		exitfail("Bad bitmap size\n");

	/* 画素データ蓄積配列をゼロクリア */
	memset(pixels, 0, sizeof(pixels));

	/* 読み込む最初の画素(MAX_HEIGHT-1行目)までシーク */
	if (fseek(pbmpfile, offbits + 3 * bmp_width * (bmp_height - MAX_HEIGHT),
		  SEEK_SET) < 0)
		exitfail_errno("fseek");
	/* 最終行から順に画素データを蓄積 */
	for (dy = DISP_HEIGHT - 1; dy >= 0; dy--)
		for (py = PIXEL_HEIGHT - 1; py >= 0; py--) {
			/* 画素データを1行分読み込む */
			if (fread(buf, sizeof(buf), 1, pbmpfile) < 1)
				exitfail_errno("fread");
			/* 画素データを加算していく */
			for (dx = 0; dx < DISP_WIDTH; dx++)
				for (px = 0; px < PIXEL_WIDTH; px++) {
					ptmp = buf[PIXEL_WIDTH * dx + px];
					pixels[dy][dx].r += (uint16_t)ptmp[2];
					pixels[dy][dx].g += (uint16_t)ptmp[1];
					pixels[dy][dx].b += (uint16_t)ptmp[0];
				}
			/* 次行先頭画素までシーク */
			if (fseek(pbmpfile, 3 * bmp_width - sizeof(buf),
				  SEEK_CUR) < 0)
				exitfail_errno("fseek");
		}

	/* ビットマップファイルクローズ */
	fclose(pbmpfile);

	/* 先頭行から順にキャラクタを描画していく */
	for (dy = 0; dy < DISP_HEIGHT; dy++) {
		printf("\n");
		for (dx = 0; dx < DISP_WIDTH; dx++) {
			/* 境界値と比較して
			   色付けエスケープシーケンスとキャラクタを出力 */
			printf("\x1b[3%dm\x1b[4%dm%c",
			       ((pixels[dy][dx].r > BOUNDARY_FG) ? 1 : 0) |
			       ((pixels[dy][dx].g > BOUNDARY_FG) ? 2 : 0) |
			       ((pixels[dy][dx].b > BOUNDARY_FG) ? 4 : 0),
			       ((pixels[dy][dx].r > BOUNDARY_BG) ? 1 : 0) |
			       ((pixels[dy][dx].g > BOUNDARY_BG) ? 2 : 0) |
			       ((pixels[dy][dx].b > BOUNDARY_BG) ? 4 : 0),
			       ((dx ^ dy) & 1) ? '_' : '/');
		}
		/* 色付けをクリアするエスケープシーケンスを出力 */
		printf("\x1b[0m");
	}
	/* 入力があるまで一時停止 */
	getchar();

	return 0;
}
