#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

#define BUF_SIZE  256

#define BASENAME(p)   ((strrchr((p), '/') ? : ((p) - 1)) + 1)

/**
 * main関数
 * @param argc 引数なしの場合はusage表示のみ
 * @param argv 第1引数として接続待ちポートを指定
 * @return exit値
 */
int main(int argc, char *argv[])
{
	in_port_t listen_port;
	static int listen_fd, accept_fd; /* ソケットファイルディスクリプタ */
	struct sockaddr_in server_addr, client_addr;
	socklen_t addr_len;
	char message[] = "Hello!\r\n";
	fd_set fds_org, rdfds, wrfds, *prdfds, *pwrfds;
	int nfds;
	char buf[BUF_SIZE];
	ssize_t ret, len, wrlen;

	exitfail_init();

	/* 引数が指定されなかった場合、usage表示して終了 */
	if (argc < 2) {
		printf("Usage: %s <port>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	ret = strtoul(argv[1], NULL, 10);
	if (ret == LONG_MIN || ret == LONG_MAX)
		exitfail_errno("strtol");
	if (ret < 49152 || 65535 < ret)
		exitfail("Specify the port 49152-65535\n");
	listen_port = ret;

	/* 接続待ち用のソケットを作成 */
	listen_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (listen_fd < 0)
		exitfail_errno("open");

	/* アドレスとポートを割り当て */
	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sin_family = AF_INET; /* IPv4インターネットプロトコル */
	server_addr.sin_addr.s_addr = INADDR_ANY; /* 任意のアドレス */
	server_addr.sin_port = htons(listen_port); /* 接続を待つポート */
	addr_len = sizeof(server_addr);
	ret = bind(listen_fd, (struct sockaddr *)&server_addr, addr_len);
	if (ret < 0)
		exitfail_errno("bind");

	/* クライアントからの接続を待つ */
	ret = listen(listen_fd, SOMAXCONN);
	if (ret < 0)
		exitfail_errno("listen");

	/* クライアントからの接続を受け付けて、入出力用のソケットを作成 */
	accept_fd = accept(listen_fd, (struct sockaddr *)&client_addr,
			   &addr_len);
	if (accept_fd < 0)
		exitfail_errno("accept");

	/* これ以上の接続を受け付けないため、接続待ち用ソケットを破棄 */
	close(listen_fd);

	/* メッセージを送信 */
	ret = write(accept_fd, message, strlen(message));
	if (ret < 0)
		exitfail_errno("write");

	/* selectのため、ソケットの設定されたfdセットを作成しておく */
	FD_ZERO(&fds_org);
	FD_SET(accept_fd, &fds_org);
	nfds = accept_fd + 1;

	len = 0;
	/* 無限ループ */
	for (; ; ) {
		/* バッファに空きがある場合、読み込み可能を待つ */
		if (len < BUF_SIZE) {
			rdfds = fds_org;
			prdfds = &rdfds;
		}
		else
			prdfds = NULL;
		/* バッファにデータがある場合、書き込み可能を待つ */
		if (len > 0) {
			wrfds = fds_org;
			pwrfds = &wrfds;
		}
		else
			pwrfds = NULL;
		/* 読み書きが可能になるまで待つ */
		ret = select(nfds, prdfds, pwrfds, NULL, NULL);
		if (ret < 0) {
			/* シグナル発生時はリトライ */
			if (errno == EINTR)
				continue;
			exitfail_errno("select");
		}

		/* 書き込み可能になった */
		if (pwrfds && FD_ISSET(accept_fd, pwrfds)) {
			/* ソケットに書き込み */
			ret = write(accept_fd, buf, len);
			if (ret < 0) {
				/* シグナル発生時はリトライ */
				if (errno == EINTR)
					continue;
				exitfail_errno("write");
			}
			wrlen = ret;

			/* 書き込んだ分を捨てて、残りデータを前にずらす */
			if (wrlen < len)
				memmove(buf, buf + wrlen, len - wrlen);
			len -= wrlen;
		}

		/* 読み込み可能になった */
		if (prdfds && FD_ISSET(accept_fd, prdfds)) {
			/* ソケットから読み込み */
			ret = read(accept_fd, buf + len, BUF_SIZE - len);
			if (ret <= 0) {
				if (ret == 0)
					break; /* 終了 */
				/* シグナル発生時はリトライ */
				if (errno == EINTR)
					continue;
				exitfail_errno("read");
			}
			len += ret;
		}
	}

	/* 入出力用のソケットを破棄 */
	close(accept_fd);

	return EXIT_SUCCESS;
}
